using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.IO;
using System.Threading;

namespace MandelbrotLib
{
	public class ComplexNumber
	{
		private double r; // Reálná část komplexního čísla.
		private double i; // Imaginární část komplexního čísla.
		
		public double R
		{
			get { return this.r; }
			set { this.r = value; }
		}

		public double I
		{
			get { return this.i; }
			set { this.i = value; }
		}

		public ComplexNumber ()
		{
			this.R = 0;
			this.I = 0;
		}

		public ComplexNumber (double r, double i)
		{
			this.R = r;
			this.I = i;
		}

		public ComplexNumber (ComplexNumber num)
		{
			this.R = num.R;
			this.I = num.I;
		}

		// Vrátí komplexně sdružené číslo k tomuto číslu.
		public static ComplexNumber ComplexConjugate (ComplexNumber num)
		{
			return new ComplexNumber (num.R, -num.I);
		}

		// Tohle je unární - (jinak řečeno, vytvoří opačné komplexní číslo).
		public static ComplexNumber operator - (ComplexNumber num)
		{
			return new ComplexNumber (-num.R, -num.I);
		}

		// Tohle je binární -.
		public static ComplexNumber operator - (ComplexNumber a, ComplexNumber b)
		{
			return new ComplexNumber (a.R - b.R, a.I - b.I);
		}

		// Tohle je binární +.
		public static ComplexNumber operator + (ComplexNumber a, ComplexNumber b)
		{
			return new ComplexNumber (a.R + b.R, a.I + b.I);
		}

		// Tohle je binární *.
		public static ComplexNumber operator * (ComplexNumber a, ComplexNumber b)
		{
			return new ComplexNumber (a.R * b.R - a.I * b.I, a.R * b.I + a.I * b.R);
		}

		// Tohle je binární /.
		public static ComplexNumber operator / (ComplexNumber a, ComplexNumber b)
		{
			// Dělení probíhá pomocí triku, kdy celký zlomek vynásobím
			// číslem komplexně sdruženým k b. Tím ve jmenovateli dostanu
			// součet dvou reálných čísel a v čitateli získám součin dvou
			// komplexních čísel, což už je možné spočítat.
			double denomination = b.R * b.R + b.I * b.I;
			ComplexNumber numerator = a * ComplexNumber.ComplexConjugate (b);
			return new ComplexNumber (numerator.R / denomination, numerator.I / denomination);
		}

		// Tohle je (zatím) celočíselná mocnina.
		public static ComplexNumber Pow (ComplexNumber num, int exponent)
		{
			if (exponent < 2)
			{
				return new ComplexNumber (num);
			}
			return new ComplexNumber (num * ComplexNumber.Pow (num, exponent - 1));
		}

		// Tohle je absolutní hodnota.
		public static double Abs (ComplexNumber num)
		{
			return Math.Sqrt (num.R * num.R + num.I * num.I);
		}

		// Přepíšeme si metodu toString.
		public override string ToString ()
		{
			//return string.Format ("[ComplexNumber: R={0}, I={1}]", R, I);
			return string.Format ("({0} + {1}i)", this.R, this.I);
		}
	}
	
	class RGB
	{
		private int r = 0;
		private int g = 0;
		private int b = 0;

		public int R
		{
			get { return this.r; }
			set {
				if ((value >= 0) && (value <= 255)) {
					this.r = value;
				}
			}
		}

		public int G
		{
			get { return this.g; }
			set {
				if ((value >= 0) && (value <= 255)) {
					this.g = value;
				}
			}
		}

		public int B
		{
			get { return this.b; }
			set {
				if ((value >= 0) && (value <= 255)) {
					this.b = value;
				}
			}
		}

		public RGB (int p_r, int p_g, int p_b)
		{
			this.R = p_r;
			this.G = p_g;
			this.B = p_b;
		}
	}
	
	public class Palette
	{
		private SortedList<int, RGB> rgb;
		private bool special = false;
		private string file = "";
		private string[] specialStr = new string[] { MandelbrotSet.SPECCOL,
			MandelbrotSet.SPECCOL, MandelbrotSet.SPECCOL };

		public bool Special
		{
			get { return this.special; }
		}

		public Palette (string p_file, bool p_special)
		{
			this.file = p_file;
			this.special = p_special;
		}

		private RGB parseLine (string line, int numberOfLine)
		{
			string[] values = line.Split (' ');
			int[] color = new int[3];
			
			try
			{
				color[0] = int.Parse (values[0]);
				color[1] = int.Parse (values[1]);
				color[2] = int.Parse (values[2]);
			}
			catch
			{
				color[0] = 0;
				color[1] = 0;
				color[2] = 0;
				Console.WriteLine ("Parsing error on line {0}. Setting this" +
					" RGB value to (0, 0, 0).", numberOfLine);
			}
			return new RGB (color[0], color[1], color[2]);
		}

		public void LoadFromFile ()
		{
			if (File.Exists (this.file))
			{
				try
				{
					this.rgb = new SortedList<int, RGB> ();
					using (StreamReader sr = File.OpenText (this.file))
					{
						int num = 0;
						while (!sr.EndOfStream)
						{
							num++;
							string line = sr.ReadLine ();
							this.rgb.Add (num - 1, this.parseLine (line, num));
						}
					}
				}
				catch (Exception e)
				{
					Console.WriteLine ("Error!");
					Console.WriteLine (e.Message);
					Console.WriteLine ("Own palette will not be used.");
					this.special = true;
				}
			}
			else
			{
				Console.WriteLine ("Error!");
				Console.WriteLine ("File \"{0}\" does not exists.", this.file);
				Console.WriteLine ("Own palette will not be used.");
				this.special = true;
			}
		}

		public void SetSpecial (string r, string g, string b)
		{
			this.specialStr[0] = r;
			this.specialStr[1] = g;
			this.specialStr[2] = b;
		}

		public Color GetColor (int i)
		{
			if (this.special)
			{
				int r, g, b;
				
				if (this.specialStr[0] == "+")
				{
					r = i;
				}
				else if (this.specialStr[0] == "-")
				{
					r = 255 - i;
				}
				else
				{
					r = int.Parse (this.specialStr[0]);
				}
				
				if (this.specialStr[1] == "+")
				{
					g = i;
				}
				else if (this.specialStr[1] == "-")
				{
					g = 255 - i;
				}
				else
				{
					g = int.Parse (this.specialStr[1]);
				}
				
				if (this.specialStr[2] == "+")
				{
					b = i;
				}
				else if (this.specialStr[2] == "-")
				{
					b = 255 - i;
				}
				else
				{
					b = int.Parse (this.specialStr[2]);
				}
				return Color.FromArgb (r, g, b);
			}
			else
			{
				int c = i;
				if (c >= this.rgb.Count)
				{
					while ((c -= this.rgb.Count) >= this.rgb.Count)
						;
				}
				RGB t_rgb = new RGB (0, 0, 0);
				if (this.rgb.TryGetValue (c, out t_rgb))
				{
					return Color.FromArgb (t_rgb.R, t_rgb.G, t_rgb.B);
				}
				else
				{
					return Color.Black;
				}
			}
		}
	}
	
	public static class MyDoubleExtension
	{
		public static bool AlmostEqual (this double d1, double d2, double precision)
		{
			// To s tím this je zajímavé, v hlavním kódu se to potom použije
			// takto: double1.AlmostEqual (d2, precision); Super :)
			return (Math.Abs (d1 - d2) <= precision);
		}
	}
	
	public delegate void bitmapDrawer();
	public delegate void onBitmapDrawed(object sender, Bitmap bmp);
	
	public class MandelbrotSet
	{
		// Konstanty. Nepoužívám const, ale static readonly, abych při změně
		// těchto hodnot nemusel rekompilovat i závislé programy.
		public static readonly int MAXITER = 255;
		public static readonly double WIDTHX = 2.0;
		public static readonly double WIDTHY = 2.0;
		public static readonly int EXPONENT = 2;
		public static readonly double SCALE = 1.0;
		public static readonly double POSR = 0.0;
		public static readonly double POSI = 0.0;
		public static readonly int WIDTH = 512;
		public static readonly int HEIGHT = 512;
		public static readonly string OUTPUTFILE = "out.png";
		public static readonly bool OWNPALETTE = false;
		public static readonly string SPECCOL = "+";

		private int maxIter = MAXITER; // Maximální počet iterací.
		
		// Výřez obrazce.
		private double minX;
		private double maxX;
		private double minY;
		private double maxY;

		// Absolutní hranice obrazce v komplexní rovině.
		private double widthX = WIDTHX;
		private double widthY = WIDTHY;

		private int exponent = EXPONENT; // Exponent, použitý pro výpočet.
		private double scale = SCALE; // Měřítko.
		
		private ComplexNumber pos = new ComplexNumber (POSR, POSI); // Posun obrázku.
		
		private int width = WIDTH; // Výchozí šířka obrázku.
		private int height = HEIGHT; // Výchozí výška obrázku.
		
		private Palette palette; // Paleta.
		
		private Bitmap bmp;
		
		private bitmapDrawer bd;
		
		public event onBitmapDrawed OnBitmapDrawed;

		public MandelbrotSet (int p_width, int p_height, double p_widthX,
			double p_widthY, double p_scale, ComplexNumber p_pos, int p_exponent,
			int p_maxIter, Palette p_palette)
		{
			this.width = p_width;
			this.height = p_height;
			// Nesmí být větší než 2.0.
			this.widthX = (p_widthX < this.widthX) ? p_widthX : this.widthX;
			this.widthY = (p_widthY < this.widthY) ? p_widthY : this.widthY;
			this.scale = p_scale;
			if ((p_pos.R >= -this.widthX) && (p_pos.R <= this.widthX) &&
				(p_pos.I >= -this.widthY) && (p_pos.I <= this.widthY))
			{
				this.pos = p_pos;
			}
			this.exponent = (p_exponent > this.exponent) ? p_exponent : this.exponent;
			this.maxIter = (p_maxIter < this.maxIter) ? p_maxIter : this.maxIter;
			this.palette = p_palette;
			this.computeCorners ();
			bmp = new Bitmap (this.width, this.height, PixelFormat.Format24bppRgb);
			this.bd = computeMandelbrot;
		}

		// Spočte hranice výřezu a měřítko.
		private void computeCorners ()
		{
			this.minX = this.pos.R - this.widthX * this.scale;
			this.minY = this.pos.I - this.widthY * this.scale;
			this.maxX = this.pos.R + this.widthX * this.scale;
			this.maxY = this.pos.I + this.widthY * this.scale;
		}

		// Vykreslí Mandelbrotovu množinu.
		private void computeMandelbrot ()
		{
			int i;
			int x, y;
			ComplexNumber c0 = new ComplexNumber (this.minX, this.minY);
			ComplexNumber c = new ComplexNumber ();
			ComplexNumber z = new ComplexNumber ();
			
			for (y = 0; y < this.height; y++)
			{
				c0.R = this.minX;
				for (x = 0; x < this.width; x++)
				{
					c.R = c0.R;
					c.I = c0.I;
					z.R = 0;
					z.I = 0;
					for (i = 0; i < this.maxIter; i++)
					{
						z = ComplexNumber.Pow (z, this.exponent) + c;
						if (ComplexNumber.Abs (z) > 2.0)
						{
							break;
						}
					}
					this.bmp.SetPixel (x, y, this.palette.GetColor (i));
					c0.R += (this.maxX - this.minX) / this.width;
				}
				c0.I += (this.maxY - this.minY) / this.height;
			}
		}
		
		public void BeginDraw ()
		{
			this.bd.BeginInvoke (onEnd, this.bd);
		}
		
		private void onEnd (IAsyncResult ir)
		{
			if (OnBitmapDrawed != null)
			{
				OnBitmapDrawed (this, this.bmp);
			}
		}
	}
}

